package ru.dbotthepony.mc.otm.network

import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.readImpreciseFraction
import ru.dbotthepony.mc.otm.core.writeImpreciseFraction
import ru.dbotthepony.mc.otm.readItem
import ru.dbotthepony.mc.otm.writeItem
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream
import kotlin.properties.ReadWriteProperty
import kotlin.reflect.KProperty

interface FieldAccess<V> {
	fun read(): V
	fun write(value: V)
}

fun interface FieldGetter<V> {
	fun invoke(field: FieldAccess<V>): V
}

fun interface FieldSetter<V> {
	fun invoke(value: V, field: FieldAccess<V>, setByRemote: Boolean)
}

class NetworkValueCodec<V>(
	val reader: (stream: DataInputStream) -> V,
	val writer: (stream: DataOutputStream, value: V) -> Unit,
	val copier: ((value: V) -> V) = { it },
	val comparator: ((a: V, b: V) -> Boolean) = { a, b -> a == b }
)

val BooleanValueCodec = NetworkValueCodec(DataInputStream::readBoolean, DataOutputStream::writeBoolean)
val IntValueCodec = NetworkValueCodec(DataInputStream::readInt, DataOutputStream::writeInt)
val ItemStackValueCodec = NetworkValueCodec(DataInputStream::readItem, DataOutputStream::writeItem, ItemStack::copy, ItemStack::isSameItemSameTags)
val ImpreciseFractionValueCodec = NetworkValueCodec(DataInputStream::readImpreciseFraction, DataOutputStream::writeImpreciseFraction)

class FieldSynchronizer {
	private val fields = ArrayList<Field<*>>()
	private val observers = ArrayList<Field<*>>()
	private val dirtyFields = ArrayList<Field<*>>()

	fun bool(
		value: Boolean = false,
		getter: FieldGetter<Boolean>? = null,
		setter: FieldSetter<Boolean>? = null,
	): Field<Boolean> {
		return Field(value, BooleanValueCodec, getter, setter)
	}

	fun int(
		value: Int = 0,
		getter: FieldGetter<Int>? = null,
		setter: FieldSetter<Int>? = null,
	): Field<Int> {
		return Field(value, IntValueCodec, getter, setter)
	}

	fun fraction(
		value: ImpreciseFraction = ImpreciseFraction.ZERO,
		getter: FieldGetter<ImpreciseFraction>? = null,
		setter: FieldSetter<ImpreciseFraction>? = null,
	): Field<ImpreciseFraction> {
		return Field(value, ImpreciseFractionValueCodec, getter, setter)
	}

	fun item(
		value: ItemStack = ItemStack.EMPTY,
		getter: FieldGetter<ItemStack>? = null,
		setter: FieldSetter<ItemStack>? = null,
	): Field<ItemStack> {
		return Field(value, ItemStackValueCodec, getter, setter, isObserver = true)
	}

	inner class Field<V>(
		private var value: V,
		private val dispatcher: NetworkValueCodec<V>,
		private val getter: FieldGetter<V>? = null,
		private val setter: FieldSetter<V>? = null,
		isObserver: Boolean = false,
	) : ReadWriteProperty<Any, V> {
		private var remote: V = dispatcher.copier.invoke(value)

		val id = fields.size + 1

		init {
			fields.add(this)

			if (isObserver) {
				observers.add(this)
			}
		}

		private var isDirty = false

		private val access = object : FieldAccess<V> {
			override fun read(): V {
				return value
			}

			override fun write(value: V) {
				if (!isDirty && !dispatcher.comparator.invoke(remote, value)) {
					dirtyFields.add(this@Field)
					isDirty = true
				}

				this@Field.value = value
			}
		}

		fun observe() {
			if (!isDirty && !dispatcher.comparator.invoke(remote, value)) {
				dirtyFields.add(this)
				isDirty = true
			}
		}

		fun getValue(): V {
			val getter = this.getter

			if (getter != null) {
				return getter.invoke(access)
			}

			return value
		}

		override fun getValue(thisRef: Any, property: KProperty<*>): V {
			return getValue()
		}

		fun setValue(value: V) {
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, false)
				return
			}

			if (this.value == value) {
				return
			}

			if (!isDirty && !dispatcher.comparator.invoke(remote, value)) {
				dirtyFields.add(this)
				isDirty = true
			}

			this.value = value
		}

		override fun setValue(thisRef: Any, property: KProperty<*>, value: V) {
			setValue(value)
		}

		fun markDirty() {
			if (!isDirty) {
				dirtyFields.add(this)
				isDirty = true
			}
		}

		fun write(stream: DataOutputStream) {
			stream.write(id)
			dispatcher.writer.invoke(stream, value)
			isDirty = false
			remote = dispatcher.copier.invoke(value)
		}

		fun read(stream: DataInputStream) {
			val value = dispatcher.reader.invoke(stream)
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
				return
			}

			this.value = value
		}
	}

	fun invalidate() {
		for (field in fields) {
			field.markDirty()
		}
	}

	fun collectNetworkPayload(): FastByteArrayOutputStream? {
		if (observers.isNotEmpty()) {
			for (field in observers) {
				field.observe()
			}
		}

		if (dirtyFields.isEmpty()) {
			return null
		}

		val stream = FastByteArrayOutputStream()
		val dataStream = DataOutputStream(stream)

		for (field in dirtyFields) {
			field.write(dataStream)
		}

		dirtyFields.clear()
		dataStream.write(0)

		return stream
	}

	fun applyNetworkPayload(stream: DataInputStream): Int {
		var fieldId = stream.read()
		var i = 0

		while (fieldId != 0) {
			val field = fields.getOrNull(fieldId - 1) ?: throw IndexOutOfBoundsException("Invalid field id $fieldId")
			field.read(stream)
			fieldId = stream.read()
			i++
		}

		return i
	}

	fun applyNetworkPayload(stream: InputStream): Int {
		return applyNetworkPayload(DataInputStream(stream))
	}
}
